#include "KmerCounter.hpp"
#include <iostream>
#include <sstream>
#include <algorithm>
#include <time.h>
#include "stacktrace.hpp"
#include <fstream>
#include <math.h>
#include "sequenceUtil.hpp"

KmerCounter::KmerCounter(unsigned int kmer_length) {

    if (kmer_length > 32) {
        throw(stacktrace() + "\n\nKmer length exceeds max of 32");  // STORING KMERS AS 64-bit INTEGERS, 2-bit ENCODING.  CANNOT GO HIGHER THAN 32 BASES
    }
    
    this->_kmer_length = kmer_length;
    
    
#ifdef __GOOGLE__
    
    kmer_int_type_t delete_val = 0xFFFFFFFFFFFFFFFFll;
    
    _kmer_counter.set_deleted_key(delete_val);
    
    
#endif
    
}

void KmerCounter::add_sequence(string& sequence, bool DS_mode, unsigned int cov) {
    
    for (unsigned int i = 0; i <= sequence.length() - _kmer_length; i++) {
        
        string kmer = sequence.substr(i, _kmer_length);
        
        // no storing of kmers containing non-GATC characters
        if (contains_non_gatc(kmer)) { 
            continue;
        }
        
        kmer_int_type_t kmer_val = kmer_to_intval(kmer);
        
#pragma omp critical (HashMap)
        {
            add_kmer(kmer_val, cov);
            
            
            if (DS_mode) {
                kmer_int_type_t rev_kmer_val = revcomp_val(kmer_val, _kmer_length);
                if (kmer_val != rev_kmer_val) {
                    add_kmer(rev_kmer_val, cov);
                }
                
            }
        }
    }
    
	
}


unsigned int KmerCounter::get_kmer_length() {
    return(_kmer_length);
}

unsigned long KmerCounter::size() {
    return(_kmer_counter.size());
}


void KmerCounter::describe_kmers() {
    
    Kmer_counter_map_iterator it;
    
    for (it = _kmer_counter.begin(); it != _kmer_counter.end(); it++) {
        
        kmer_int_type_t kmer_val = it->first;
        
        string kmer = decode_kmer_from_intval(kmer_val, _kmer_length);
        
        cout << describe_kmer(kmer) << endl;
        
    }
    
}


bool KmerCounter::prune_kmer_extensions( float min_ratio_non_error) {
    
    Kmer_counter_map_iterator it;
    
    Kmer_visitor deletion_list;
    
    for (it = _kmer_counter.begin(); it != _kmer_counter.end(); it++) {
        
        kmer_int_type_t kmer_val = it->first;
        int count = it->second;
        if (count == 0) {
            continue;
        }
        
        string kmer = decode_kmer_from_intval(kmer_val, _kmer_length);
        
        vector<string> candidates = get_forward_kmer_candidates(kmer);
        
        vector<string> candidates_exist;
        
        int dominant_count = 0;
        
        for (int i = 0; i < 4; i++) {
            string kmer = candidates[i];
            if (kmer_exists(kmer)) {
                int count = get_kmer_count(kmer);
                if (dominant_count == 0) {
                    dominant_count = count;
                }
                else if (dominant_count > 0 && (float) count/dominant_count < min_ratio_non_error) {
                    kmer_int_type_t kmer_val = kmer_to_intval(kmer);
                    
                    deletion_list[ kmer_val ] = true;
                    _kmer_counter[ kmer_val ] = 0; // disable when encountered in further iterations.
                    
                }
            }
        }
        
    }
    
    if (deletion_list.size() > 0) {
        
        Kmer_visitor_iterator it;
        
        for (it = deletion_list.begin(); it != deletion_list.end(); it++) {
            prune_kmer(it->first);
        }
        
        return(true);
    }
    else {
        return(false);
    }
}

void KmerCounter::dump_kmers_to_file(string& filename) {
    
    ofstream outfile (filename.c_str());
    
    if (! outfile.is_open()) {
        throw (stacktrace() + " Error, cannot write to file: " + filename);
    }
    
    
    
    Kmer_counter_map_iterator it;
    
    for (it = _kmer_counter.begin(); it != _kmer_counter.end(); it++) {
        
        kmer_int_type_t kmer_val = it->first;
        unsigned int count = it->second;
        
        string kmer = decode_kmer_from_intval(kmer_val, _kmer_length);
        
        outfile << kmer << " " << count << endl;
        
    }
    
    outfile.close();
    
}

bool KmerCounter::kmer_exists(string kmer) {
    
    kmer_int_type_t kmer_val = kmer_to_intval(kmer);
    
    return(kmer_exists(kmer_val));
    
    
}

bool KmerCounter::kmer_exists(kmer_int_type_t kmer_val) {
    
    Kmer_counter_map_iterator it = _kmer_counter.find(kmer_val);
    
    if (it != _kmer_counter.end()) {
        return(true);
    }
    else {
        return(false);
    }
}


bool KmerCounter::prune_kmer(string kmer) {
    
    if (kmer_exists(kmer)) {
        
        kmer_int_type_t kmer_val = kmer_to_intval(kmer);
        return(prune_kmer(kmer_val));
    }
    
    return(false); // no kmer in map to remove.
}

bool KmerCounter::prune_kmer(kmer_int_type_t kmer_val) {
    
    if (kmer_exists(kmer_val)) {
        
        Kmer_counter_map_iterator it = _kmer_counter.find(kmer_val);
        
#pragma omp critical (HashMap)
        {
            _kmer_counter.erase(it);
        }
        
        return(true);
    }
    return(false);
}




unsigned int KmerCounter::get_kmer_count(string kmer) {
    
    if (! kmer_exists(kmer)) {
        //throw(stacktrace() + "\n\nError, kmer " + kmer + " doesn't exist");
        return(0);
    }
    
    kmer_int_type_t kmer_val = kmer_to_intval(kmer);
    
    Kmer_counter_map_iterator it = _kmer_counter.find(kmer_val);
    
    return(it->second);
}


unsigned int KmerCounter::get_kmer_count(kmer_int_type_t kmer_val) {
    
    string kmer = decode_kmer_from_intval(kmer_val, _kmer_length);  
    
    return(get_kmer_count(kmer));
}


string KmerCounter::get_kmer_string(kmer_int_type_t kmer_val) {
    
    string kmer = decode_kmer_from_intval(kmer_val, _kmer_length);  
    
    return(kmer);
}


kmer_int_type_t KmerCounter::get_kmer_intval(string kmer) {
    
    kmer_int_type_t kmer_val = kmer_to_intval(kmer);
    
    return(kmer_val);
}


bool KmerCounter::add_kmer (kmer_int_type_t kmer_val, unsigned int count) {
    
    if (kmer_exists(kmer_val)) {
        int c = get_kmer_count(kmer_val);
        _kmer_counter[kmer_val] = c + count;
    }
    else {
        _kmer_counter[kmer_val] = count;
    }
    
    
    return(true);
    
}



bool KmerCounter::add_kmer(string kmer, unsigned int count) {
    
    // no storing of kmers containing non-GATC characters
    if (contains_non_gatc(kmer)) { 
        return(false);
    }
    
    kmer_int_type_t kmer_val = kmer_to_intval(kmer);
    
    return(add_kmer(kmer_val, count));
}


vector<string> KmerCounter::get_forward_kmer_candidates(string& seed_kmer) {
    
    vector<string> candidates = get_forward_kmer_candidates_unsorted(seed_kmer);
    
    Sort_kmer_by_count_desc sorter (this->_kmer_counter);
    sort (candidates.begin(), candidates.end(), sorter);
    
    return(candidates);
}



vector<string> KmerCounter::get_forward_kmer_candidates_unsorted(string& seed_kmer) {
    
    string forward_prefix = seed_kmer.substr(1, seed_kmer.length() - 1);
    
    vector<string> candidates;
    
    candidates.push_back(forward_prefix + "G");
    candidates.push_back(forward_prefix + "A");
    candidates.push_back(forward_prefix + "T");
    candidates.push_back(forward_prefix + "C");
    
    return(candidates);
}





vector<string> KmerCounter::get_reverse_kmer_candidates(string& seed_kmer) {
    
    string reverse_suffix = seed_kmer.substr(0, seed_kmer.length() - 1);
    
    vector<string> candidates = get_reverse_kmer_candidates_unsorted(seed_kmer);
    
    Sort_kmer_by_count_desc sorter (this->_kmer_counter);
    sort (candidates.begin(), candidates.end(), sorter);
    
    return(candidates);
}


vector<string> KmerCounter::get_reverse_kmer_candidates_unsorted(string& seed_kmer) {
    
    string reverse_suffix = seed_kmer.substr(0, seed_kmer.length() - 1);
    
    vector<string> candidates;
    
    candidates.push_back("G" + reverse_suffix);
    candidates.push_back("A" + reverse_suffix);
    candidates.push_back("T" + reverse_suffix);
    candidates.push_back("C" + reverse_suffix);
    
    return(candidates);
}





string KmerCounter::describe_kmer(string& kmer) {
    
    int kmer_count = get_kmer_count(kmer);
    
    
    // get counts for reverse candidates:
    vector<string> reverse_candidates = get_reverse_kmer_candidates_unsorted(kmer);
    
    string rG_kmer = reverse_candidates[0];
    string rA_kmer = reverse_candidates[1];
    string rT_kmer = reverse_candidates[2];
    string rC_kmer = reverse_candidates[3];
    
    int rG_count = (kmer_exists(rG_kmer)) ? get_kmer_count(rG_kmer) : 0;
    int rA_count = (kmer_exists(rA_kmer)) ? get_kmer_count(rA_kmer) : 0;
    int rT_count = (kmer_exists(rT_kmer)) ? get_kmer_count(rT_kmer) : 0;
    int rC_count = (kmer_exists(rC_kmer)) ? get_kmer_count(rC_kmer) : 0;
    
    
    // get counts for forward candidates:
    
    vector<string> forward_candidates = get_forward_kmer_candidates_unsorted(kmer);
    
    string fG_kmer = forward_candidates[0];
    string fA_kmer = forward_candidates[1];
    string fT_kmer = forward_candidates[2];
    string fC_kmer = forward_candidates[3];
    
    int fG_count = (kmer_exists(fG_kmer)) ? get_kmer_count(fG_kmer) : 0;
    int fA_count = (kmer_exists(fA_kmer)) ? get_kmer_count(fA_kmer) : 0;
    int fT_count = (kmer_exists(fT_kmer)) ? get_kmer_count(fT_kmer) : 0;
    int fC_count = (kmer_exists(fC_kmer)) ? get_kmer_count(fC_kmer) : 0;
    
    
    stringstream descr;
    
    descr << "G:" << rG_count << " " 
          << "A:" << rA_count << " "
          << "T:" << rT_count << " " 
          << "C:" << rC_count << " "
		
          << "\t" << kmer << ":" << kmer_count << "\t"
		
          << "G:" << fG_count << " "
          << "A:" << fA_count << " "
          << "T:" << fT_count << " " 
          << "C:" << fC_count;
    
    return(descr.str());
    
    
}


void KmerCounter::prune_kmers_min_count(unsigned int min_count) {
    
    Kmer_counter_map_iterator it;
    
    vector<kmer_int_type_t> kmers_to_purge;
    
    for (it = _kmer_counter.begin(); it != _kmer_counter.end(); it++) {
        
        kmer_int_type_t kmer_val = it->first;
        
        unsigned int count = it->second;
        
        if (count < min_count) {
            kmers_to_purge.push_back(kmer_val);
        }
        
    }
    
    for (unsigned int i = 0; i < kmers_to_purge.size(); i++) {
        
        prune_kmer( kmers_to_purge[i] );
    }
    
    return;
}



void KmerCounter::prune_kmers_min_entropy(float min_entropy) {
    
    Kmer_counter_map_iterator it;
    
    vector<kmer_int_type_t> kmers_to_purge;
    
    for (it = _kmer_counter.begin(); it != _kmer_counter.end(); it++) {
        
        kmer_int_type_t kmer_val = it->first;
        
        string kmer = get_kmer_string(kmer_val);
        
        float entropy = compute_entropy(kmer);
        
        if (entropy < min_entropy) {
            kmers_to_purge.push_back(kmer_val);
        }
        
    }
    
    for (unsigned int i = 0; i < kmers_to_purge.size(); i++) {
        
        prune_kmer( kmers_to_purge[i] );
    }
    
    return;
}







const Kmer_counter_map& KmerCounter::get_kmer_counter_map() const {
    
    return (_kmer_counter);
    
}

vector<kmer_int_type_t> KmerCounter::get_kmers_sort_descending_counts() {
    
    // cerr << "Getting vec of kmers" << endl;
    
    vector<kmer_int_type_t> kmer_list;
    
    Kmer_counter_map_iterator it;
    
    for (it = _kmer_counter.begin(); it != _kmer_counter.end(); it++) {
        
        kmer_list.push_back(it->first);
    }
    
    Sort_kmer_by_count_desc sorter (this->_kmer_counter);
    
    unsigned int num_kmers = kmer_list.size();
    
    unsigned long start, end;
    
    start = time(NULL);
    
    cerr << "Sorting " << num_kmers << " kmers ...";
    
    sort(kmer_list.begin(), kmer_list.end(), sorter);
    
    end = time (NULL);
    
    long time_spent = end - start;
    
    cerr << "Done sorting " << num_kmers << " kmers, taking " << time_spent << " seconds." << endl;
    
    return(kmer_list);
}


/******************/
/*  Sort Classes */
/****************/


bool Sort_kmer_by_count_desc::operator() (const kmer_int_type_t& i, const kmer_int_type_t& j) { 
    
    if (kmer_map[i] == kmer_map[j])
        return (i > j);
	else
        return (kmer_map[i] > kmer_map[j]);
    
}

bool Sort_kmer_by_count_desc::operator() (const string& i, const string& j) {
    
    kmer_int_type_t i_val = kmer_to_intval(i);
    kmer_int_type_t j_val = kmer_to_intval(j);
    
    unsigned int count_i = 0;
    if (kmer_map.find(i_val) != kmer_map.end()) {
        count_i = kmer_map[i_val];
    }
    
    unsigned int count_j = 0;
    if (kmer_map.find(j_val) != kmer_map.end()) {
        count_j = kmer_map[j_val];
    }
    
    return(count_i > count_j);
}
