#!/usr/bin/env perl

use strict;
use warnings;

use FindBin;
use File::Basename;
use Cwd;
use Carp;
use Getopt::Long qw(:config no_ignore_case bundling);


my $usage = <<_EOUSAGE_;

##################################################################
#
# Required:
#
#  --seqType <string>  :type of reads: (fq or fa)
#
#  If paired reads:
#
#      --left  <string>    :left reads
#     --right <string>    :right reads
# 
#  If unpaired reads:
#
#      --single <string>   :single reads
#

#  --output <string>     :name of directory for output (will be created if it doesn't already exist)
#
#
#  if strand-specific data, set:
#  
#      --SS_lib_type <string>  :if paired: RF or FR,  if single: F or R
#
#  
#
#  Butterfly options:
# 
#      --run_butterfly                 :executes butterfly commands.  Do not set this if you want to spawn them on a computing grid.   
#
#      --bfly_opts                     :parameters to pass through to butterfly (see butterfly documentation) default: "--edge-thr=0.16 --stderr -V 5"
#
#      --bflyHeapSpace                 :java heap space setting for butterfly (default: 1000M) => yields command java -Xmx1000M -jar Butterfly.jar ... \$bfly_opts
#   
# Misc:
#
#  --CPU <int>               :number of CPUs to use, default: 2 
#
#  --min_contig_length <int> :minimum assembled contig length to report (def=300)
#
#  --paired_fragment_length <int>  :size of a read pair insert (def=300)
#
#  --jaccard_clip     :option, set if you have paired reads and you expect high gene density with UTR overlap (use FASTQ input files for reads).
#
#  --run_ALLPATHSLG_error_correction :runs the read error correction process built into ALLPATHSLG.
#                                     (requires ALLPATHSLG to be installed, and installation directory indicated
#                                      by the env variable 'ALLPATHSLG_BASEDIR')
#
#
#############################################################################################################



_EOUSAGE_

	;



my $advanced_options = <<__ADVANCED__;

#
#  --inchworm <string>   :start with existing inchworm assemblies (skips the inchworm step given existing output)  
#  --USE_READS <string>  :use the fasta reads already extracted


__ADVANCED__
	
	;

my $INCHWORM_DIR = "$FindBin::Bin/Inchworm";
my $CHRYSALIS_DIR = "$FindBin::Bin/Chrysalis";
my $BUTTERFLY_DIR = "$FindBin::Bin/Butterfly";


# option list:
my ($seqType, $left_file, $right_file, $single_file, $SS_lib_type, $min_contig_length,
	$paired_fragment_length, $jaccard_clip, $inchworm_file, $show_advanced_options,
	$output_directory, $run_ALLPATHSLG_error_correction_flag,
	);

# defaults:
$min_contig_length = 300;
$paired_fragment_length = 300;


my $RUN_BUTTERFLY_FLAG = 0;
my $bfly_opts = "--edge-thr=0.16 --stderr -V 5";
my $bflyHeapSpace = "1000M";

my $CPU = 2;

## ADVANCED OPTIONS:
     # todo  add some.

my $use_reads;





&GetOptions( 

    ## general opts
    "seqType=s" => \$seqType,
    "left=s" => \$left_file,
    "right=s" => \$right_file,
    "single=s" => \$single_file,
    
    "SS_lib_type=s" => \$SS_lib_type,
    
    "output=s" => \$output_directory,
    
    "min_contig_length=i" => \$min_contig_length,
    "paired_fragment_length=i" => \$paired_fragment_length,
    "jaccard_clip" => \$jaccard_clip,
    
    
    # Butterfly opts
    'run_butterfly' => \$RUN_BUTTERFLY_FLAG,
    'bfly_opts=s' => \$bfly_opts,
    'bflyHeapSpace=s' => \$bflyHeapSpace,
     
    'CPU=i' => \$CPU,
    'run_ALLPATHSLG_error_correction' => \$run_ALLPATHSLG_error_correction_flag,
    


    "show_advanced_options" => \$show_advanced_options,
    
    
    ## Advanced options
    
    "inchworm=s" => \$inchworm_file,
    "USE_READS=s" => \$use_reads,
    
    );


## Check options set:

unless ($seqType && 
		( 
		  ($left_file && $right_file) || $single_file || $use_reads) 
		) {
	die $usage;
}



unless ($output_directory) {
	$output_directory = "trinity_dir.$$";
}

if ($bflyHeapSpace !~ /^\d+M$/) {
    die "Error, bflyHeapSpace must be set to a value of format: \\d+M  (eg. 1000M)";
}


if ($run_ALLPATHSLG_error_correction_flag) {
	unless ($ENV{ALLPATHSLG_BASEDIR}) {
		die "fatal: for error-correction by ALLPATHSLG, you must have ALLPATHSLG installed and the environmental"
			. " variable 'ALLPATHSLG_BASEDIR' set. ";
	}
	
	unless ($seqType eq "fq") {
		die "Error, the error-correction by AllpathsLG requires fastq files as a starting point. ";
	}

}

$ENV{OMP_NUM_THREADS} = $CPU; ## for Inchworm


main: {

	my $start_dir = cwd();

	## create complete paths for input files:
	$left_file = &create_full_path($left_file) if $left_file;
	$right_file = &create_full_path($right_file) if $right_file;
	$single_file = &create_full_path($single_file) if $single_file;
	$inchworm_file = &create_full_path($inchworm_file) if $inchworm_file;
	$output_directory = &create_full_path($output_directory);
	
	$use_reads = &create_full_path($use_reads) if $use_reads;

	unless (-d $output_directory) {
		
		mkdir $output_directory or die "Error, cannot mkdir $output_directory";
	}
	
	chdir ($output_directory) or die "Error, cannot cd to $output_directory";
	
	my $trinity_target_fa = "";
	
	if ($use_reads) {
		$trinity_target_fa = "reads.fa";
		&process_cmd("ln -s $use_reads $trinity_target_fa");
	}
	elsif ($left_file && $right_file) {
		
		if ($run_ALLPATHSLG_error_correction_flag) {
			&process_cmd("$FindBin::Bin/util/run_AllpathsLG_error_correction.pl $left_file $right_file");
			$left_file = "$left_file.ErrCor.fq";
			$right_file = "$right_file.ErrCor.fq";
		}
		
		my ($left_SS_type, $right_SS_type);
		if ($SS_lib_type) {
			($left_SS_type, $right_SS_type) = split(//, $SS_lib_type);
		}
		
		&prep_seqs($left_file, $seqType, "left", $left_SS_type) unless (-s "left.fa");
		&prep_seqs($right_file, $seqType, "right", $right_SS_type) unless (-s "right.fa");
		
		$trinity_target_fa = "both.fa";
		&process_cmd("cat left.fa right.fa > $trinity_target_fa") unless (-s $trinity_target_fa);
		
	}
	elsif ($single_file) {
		if ($run_ALLPATHSLG_error_correction_flag) {
			&process_cmd("$FindBin::Bin/util/run_AllpathsLG_error_correction.pl $single_file");
			$single_file = "$single_file.ErrCor.fq";
		}
		
		&prep_seqs($single_file, $seqType, "single", $SS_lib_type) unless (-s "single.fa");
		$trinity_target_fa = "single.fa";
	}
	
	else {
		die "not sure what to do. "; # should never get here.
	}
	
	#################
	## Inchworm step:
	
	unless ($inchworm_file) {
		$inchworm_file = &run_inchworm($trinity_target_fa, $SS_lib_type);
		
	}
	
	if ($jaccard_clip && $left_file && $right_file) {
		$inchworm_file = &run_jaccard_clip($inchworm_file, $left_file, $right_file, $SS_lib_type);
	}
	
	
	
	##################
	## Chrysalis step:
	
	my $butterfly_cmds = &run_chrysalis($inchworm_file, $trinity_target_fa,
										$min_contig_length, $paired_fragment_length, $SS_lib_type);



	print "Inchworm and Chrysalis complete.  Butterfly commands to execute are provided here:\n"
		. $butterfly_cmds . "\n\n";
	


	if ($RUN_BUTTERFLY_FLAG) {

		my $cmd = "$FindBin::Bin/util/cmd_process_forker.pl -c $butterfly_cmds --CPU $CPU";
		&process_cmd($cmd);

		
		## capture results:
		$cmd = 'find ./chrysalis -regex ".*allProbPaths.fasta" -exec cat {} \; > Trinity.fasta';

		&process_cmd($cmd);

        
		print "\n\n";
        print "###################################################################\n";
        print "Butterfly assemblies are written to $output_directory/Trinity.fasta\n";
        print "###################################################################\n\n\n";
		
        
	}
	
	exit(0);
	
}


####
sub run_chrysalis {
	my ($inchworm_file, $reads_file,
		$min_contig_length, $paired_fragment_length, $SS_lib_type) = @_;


    my $butterfly_cmds = &create_full_path("chrysalis/butterfly_commands");
    
    my $adjusted_butterfly_cmds = "$butterfly_cmds.adj";
    
    if (! -s $butterfly_cmds) {

        my $cmd = "$CHRYSALIS_DIR/Chrysalis -i $reads_file -iworm $inchworm_file -o chrysalis "
            . " -min $min_contig_length -dist $paired_fragment_length ";
        
        if ($SS_lib_type) {
            $cmd .= " -strand 1 ";
        }
        
        $cmd .= " -butterfly $BUTTERFLY_DIR/Butterfly.jar ";
        
        &process_cmd($cmd);
    }
    
	
	unless (-s $butterfly_cmds) {
		croak "Error, chrysalis did not report butterfly commands file: $butterfly_cmds";
	}
	
    ## add additional butterfly opts.
    open (my $fh, $butterfly_cmds) or die "Error, cannot read file $butterfly_cmds";
    open (my $ofh, ">$adjusted_butterfly_cmds") or die "Error, cannot write to $adjusted_butterfly_cmds";
    while (<$fh>) {
        my $line = $_;
        chomp $line;
        $line =~ s/^java /java -Xmx$bflyHeapSpace / or die "Error, cannot modify command";
        if ($bfly_opts) {
            $line .= " $bfly_opts";
        }
        print $ofh $line . "\n";
    }
    close $ofh;
    close $fh;
    

	return($adjusted_butterfly_cmds);
	
}


####
sub run_inchworm {
	my ($reads, $strand_specific_flag) = @_;
	
	my $inchworm_outfile = "inchworm.K25.L48";

	my $inchworm_cmd = "$INCHWORM_DIR/bin/inchworm --reads $reads --run_inchworm -K 25 -L 48 --monitor 1 ";
    unless ($strand_specific_flag) {
		$inchworm_cmd .= " --DS ";
		$inchworm_outfile .= ".DS";
	}
	
	$inchworm_outfile .= ".fa";
	
	$inchworm_cmd .= " 2>monitor.out > $inchworm_outfile";
	
	&process_cmd($inchworm_cmd) unless (-s $inchworm_outfile);

	return($inchworm_outfile);
	
}

####
sub prep_seqs {
	my ($initial_file, $seqType, $file_prefix, $SS_lib_type) = @_;

	if ($seqType eq "fq") {
		# make fasta
		
		my $cmd = "$INCHWORM_DIR/util/fastQ_to_fastA.pl -I $initial_file ";
		if ($SS_lib_type && $SS_lib_type eq "R") {
			$cmd .= " --rev ";
		}
		$cmd .= "> $file_prefix.fa";
		
		&process_cmd($cmd) unless (-e "$file_prefix.fa");
	}
	elsif ($seqType eq "fa") {
		if ($SS_lib_type && $SS_lib_type eq "R") {
			my $cmd = "$INCHWORM_DIR/util/revcomp_fasta.pl $initial_file > $file_prefix.fa";
			&process_cmd($cmd) unless (-s "$file_prefix.fa");
		}
		else {
			## just symlink it here:
			my $cmd = "ln -s $initial_file $file_prefix.fa";
			&process_cmd($cmd) unless (-s "$file_prefix.fa");
		}
	}
	
	return;
}



###
sub create_full_path {
	my ($file) = @_;

	my $cwd = cwd();
	if ($file !~ m|^/|) { # must be a relative path
		$file = $cwd . "/$file";
	}
	
	return($file);
}



####
sub process_cmd {
	my ($cmd) = @_;

	print "CMD: $cmd\n";

	my $ret = system($cmd);

	if ($ret) {
		die "Error, cmd: $cmd died with ret $ret";
	}
	
	return;
}


####
sub run_jaccard_clip {
	my ($inchworm_file, $left_file, $right_file, $SS_lib_type) = @_;

	my $cmd = "$INCHWORM_DIR/bin/inchworm_transcript_splitter.pl --iworm $inchworm_file "
		. " --left_fq $left_file --right_fq $right_file ";

	if ($SS_lib_type) {
		$cmd .= " --SS_lib_type $SS_lib_type ";
	}
	
	&process_cmd($cmd);

	my $output_file = "$inchworm_file.clipped.fa";

	unless (-s $output_file) {
		croak "Error, jaccard clipping didn't produce the expected output file: $output_file";
	}

	return($output_file);
}
